from hitchstory.result import Success, Failure
from hitchstory.story_step import StoryStep
from hitchstory.utils import DEFAULT_STACK_TRACE
from hitchstory import exceptions
from hitchstory import utils
from strictyaml import Optional
from slugify import slugify
import time


class Story(object):
    def __init__(self, story_file, name, parsed_yaml, variation_of=None):
        self._story_file = story_file
        self._name = name
        self._slug = None
        self._parsed_yaml = parsed_yaml
        self._steps = []
        self._info = {}
        self._parent = None
        self._variation_of = variation_of
        if self.engine.schema.info is not None:
            for info_property in self.engine.schema.info.keys():
                key = info_property.key if isinstance(info_property, Optional) \
                    else info_property
                self._info[key] = self.data.get(key)
        self._collection = self._story_file.collection
        self.children = []

    def _unparameterized_preconditions(self):
        precondition_dict = self.parent._unparameterized_preconditions() \
            if self.parent is not None else {}
        for name, precondition in self.data.get("given", {}).items():
            precondition_dict[name] = precondition
        return precondition_dict

    @property
    def _yaml_steps(self):
        step_list = self._parent_steps
        step_list.extend(self._parsed_yaml.get('steps', []))
        return step_list

    @property
    def _parent_steps(self):
        return self.parent._yaml_steps \
            if self.parent is not None else []

    def _initialize(self):
        precondition_dict = self._unparameterized_preconditions()
        for name, precondition in precondition_dict.items():
            if utils.is_parameter(precondition):
                precondition_dict[name] = \
                    self.params[utils.parameter_name(precondition)]
            else:
                precondition_dict[name] = precondition
        self._precondition_dict = precondition_dict
        number_of_parent_steps = len(self._parent_steps)
        self._steps = [
            StoryStep(
                self, parsed_step, index, index - number_of_parent_steps, self.params
            ) for index, parsed_step in enumerate(self._yaml_steps)
        ]

    def update(self, step, kwargs):
        self._story_file.update(self, step, kwargs)

    @property
    def variation(self):
        return self._variation_of is not None

    @property
    def data(self):
        return self._parsed_yaml.data

    @property
    def based_on(self):
        return self.data['based on'] if "based on" in self.data else self._variation_of

    @property
    def story_file(self):
        return self._story_file

    @property
    def engine(self):
        return self._story_file.engine

    @property
    def filename(self):
        return self._story_file.filename

    @property
    def info(self):
        return self._info

    @property
    def child_name(self):
        return self._name

    @property
    def name(self):
        return self._name if not self.variation\
            else "{0}/{1}".format(self._variation_of, self._name)

    @property
    def slug(self):
        if self._slug is None:
            self._slug = slugify(self.name)
        return self._slug

    @property
    def parent(self):
        return self._parent

    @property
    def params(self):
        param_dict = self.parent.params \
            if self.parent is not None else {}
        for name, param in self.data.get("with", {}).items():
            param_dict[name] = param
        for name, param in self._collection._params.items():
            param_dict[name] = param
        return param_dict

    @property
    def given(self):
        return self._precondition_dict

    @property
    def steps(self):
        return self._steps

    def run_special_method(self, method, exception_to_raise, result=None):
        try:
            if result is None:
                method()
            else:
                method(result)
        except Exception as exception:
            stack_trace = DEFAULT_STACK_TRACE.current_stacktrace()

            raise exception_to_raise(
                stack_trace
            )

    def run_on_success(self):
        try:
            self.engine.on_success()
        except Exception as exception:
            self.run_special_method(self.engine.tear_down, exceptions.TearDownException)
            stack_trace = DEFAULT_STACK_TRACE.current_stacktrace()

            raise exceptions.OnSuccessException(stack_trace)

    def run_on_failure(self, result):
        try:
            self.engine.on_failure(result)
        except Exception as exception:
            self.run_special_method(self.engine.tear_down, exceptions.TearDownException)
            stack_trace = DEFAULT_STACK_TRACE.current_stacktrace()

            raise exceptions.OnFailureException(stack_trace)

    def documentation(self, template="story"):
        return utils.render_template(
            self._collection._templates,
            template,
            {"story": self, }
        )

    def play(self):
        """
        Run a story from beginning to end, time it and return
        a Result object.
        """
        start_time = time.time()
        passed = False
        caught_exception = None
        try:
            from signal import signal, SIGINT, SIGTERM, SIGHUP, SIGQUIT

            signal(SIGINT, self.engine.on_abort)
            signal(SIGTERM, self.engine.on_abort)
            signal(SIGHUP, self.engine.on_abort)
            signal(SIGQUIT, self.engine.on_abort)

            current_step = None
            self.engine._preconditions = self.given
            self.engine.story = self
            self.engine.set_up()

            for step in self.steps:
                current_step = step
                self.engine.current_step = current_step
                run_step_method = step.method()
                run_step_method()

                if hasattr(self.engine, '_aborted') and self.engine._aborted:
                    print("Aborted")
            passed = True
        except Exception as exception:
            caught_exception = exception
            if current_step is not None and current_step.expect_exception(
                self.engine, caught_exception
            ):
                failure_stack_trace = DEFAULT_STACK_TRACE.only_the_exception().current_stacktrace()
            else:
                failure_stack_trace = DEFAULT_STACK_TRACE.current_stacktrace()

        if passed:
            self.run_on_success()
            result = Success(self, time.time() - start_time)
        else:
            result = Failure(
                self,
                time.time() - start_time,
                caught_exception,
                current_step,
                failure_stack_trace,
            )
            self.run_on_failure(result)

        self.run_special_method(self.engine.tear_down, exceptions.TearDownException)
        return result

    def __repr__(self):
        return u"Story('{0}')".format(self._slug)



from hitchstory.step_method import StepMethod
from hitchstory.arguments import Arguments
from hitchstory import exceptions
from hitchstory import utils


class StoryStep(object):
    def __init__(self, story, yaml_step, index, child_index, params):
        self._yaml = yaml_step
        self._story = story
        self._index = index
        self._slug = None
        self._child_index = child_index
        if yaml_step.is_scalar():
            self.name = yaml_step.data
            self.arguments = Arguments(None, params)
        elif yaml_step.is_mapping():
            self.name = yaml_step.keys()[0].data
            self.arguments = Arguments(list(yaml_step.values())[0], params)
            StepMethod(self.step_method).revalidate(self.arguments)

    def underscore_case_name(self):
        return utils.to_underscore_style(self.name)

    def update(self, **kwargs):
        self._story.update(self, kwargs)

    @property
    def slug(self):
        return self.underscore_case_name()

    @property
    def index(self):
        return self._index

    @property
    def child_index(self):
        return self._child_index

    @property
    def yaml(self):
        return self._yaml

    def documentation(self, template=None):
        return utils.render_template(
            self._story._collection._templates,
            template if template is not None else self.slug,
            {"step": self, }
        )

    @property
    def _data(self):
        if self.arguments.single_argument:
            return {
                StepMethod(self.step_method).single_argument_name: self.arguments.data
            }
        return self.arguments.data

    def __getitem__(self, name):
        return self._data[name]

    def __contains__(self, name):
        return name in self._data

    @property
    def step_method(self):
        engine = self._story.engine
        if hasattr(engine, self.underscore_case_name()):
            attr = getattr(engine, self.underscore_case_name())
            if hasattr(attr, '__call__'):
                return attr
            else:
                raise exceptions.StepNotCallable((
                    "Step with name '{}' in {} is not a function "
                    "or a callable object, it is a {}".format(
                        self.underscore_case_name(),
                        engine.__repr__(),
                        type(attr)
                    )
                ))
        else:
            raise exceptions.StepNotFound("Step with name '{}' not found in {}.".format(
                self.underscore_case_name(),
                engine.__repr__()
            ))

    def expect_exception(self, engine, exception):
        if isinstance(exception, exceptions.Failure):
            return True

        try:
            step_method = self.step_method
        except exceptions.HitchStoryException:
            return False

        if hasattr(step_method, '_expected_exceptions'):
            return isinstance(exception, tuple(step_method._expected_exceptions))

        return False

    def method(self):
        return StepMethod(self.step_method).method(self.arguments)

    def __repr__(self):
        return u"<StoryStep('{0}')>".format(self.slug)



from hitchstory.story_collection import StoryCollection
from hitchstory.engine import validate
from hitchstory.engine import expected_exception
from hitchstory.engine import BaseEngine
from hitchstory.engine import StorySchema
from hitchstory.exceptions import HitchStoryException
from hitchstory.exceptions import Failure


__version__ = "0.7.4"



from hitchstory import utils
from strictyaml import ScalarValidator, Map
from slugify import slugify


class UnderscoreCase(ScalarValidator):
    def validate_scalar(self, chunk):
        return slugify(chunk.contents, separator='_')


class Arguments(object):
    """A null-argument, single argument or group of arguments of a hitchstory step."""

    def __init__(self, yaml_args, params):
        """Create arguments from dict (from yaml)."""
        self._params = params

        if yaml_args is None:
            self.is_none = True
            self.single_argument = False
        elif yaml_args.is_mapping():
            self.is_none = False
            self.single_argument = False
            self.yaml = yaml_args
        elif yaml_args.is_scalar():
            self.is_none = False
            self.single_argument = True
            self.yaml = yaml_args

    def parameterize(self, value):
        """
        Replace parameters with specified variables.
        """
        for name, parameter in self._params.items():
            if utils.is_parameter(value):
                if name == utils.parameter_name(value):
                    return parameter
        return value

    def validate_args(self, validator):
        """
        Validate step using StrictYAML validators specified in @validate decorators.
        """
        self.yaml.revalidate(Map(validator, key_validator=UnderscoreCase()))
        self.data = {}

        for key, value in self.yaml.items():
            self.data[key.data] = self.parameterize(value.data)

    def validate_single_argument(self, validator):
        self.yaml.revalidate(validator)
        self.data = self.parameterize(self.yaml.data)

    def validate_kwargs(self, validator):
        self.yaml.revalidate(validator)
        self.data = self.parameterize(self.yaml.data)



from hitchstory import exceptions
from strictyaml import Any, Optional
from hitchstory import utils
from functools import partial
import inspect


class StepMethod(object):
    def __init__(self, method):
        self._method = method
        if self.argspec.varargs is not None:
            raise exceptions.CannotUseVarargs(
                "Cannot use varargs (e.g. *args), can only use keyword args (**kwargs)"
            )
        if self._keywords and len(self._args) > 1:
            raise exceptions.CannotMixKeywordArgs(
                "Cannot mix keyword arguments (e.g. **kwargs) and regular args (e.g. x)"
            )

    @property
    def argspec(self):
        return inspect.getargspec(self._method)

    @property
    def _defaults(self):
        return self.argspec.defaults if self.argspec.defaults is not None else []

    @property
    def _args(self):
        return self.argspec.args if self.argspec.args is not None else []

    @property
    def _specified_validators(self):
        return self._method._validators if hasattr(self._method, '_validators') else {}

    @property
    def _keywords(self):
        return self.argspec.keywords is not None

    def arg_validator(self, name):
        return utils.YAML_Param | self._specified_validators.get(name, Any())

    @property
    def optional_args(self):
        return self.argspec.args[-len(self._defaults):]

    @property
    def required_args(self):
        return self.argspec.args[:len(self._args) - len(self._defaults)][1:]

    @property
    def single_argument_allowed(self):
        return len(self.required_args) == 0 and len(self.optional_args) > 0 \
            or len(self.required_args) == 1

    @property
    def single_argument_name(self):
        return self.argspec.args[1]

    @property
    def mapping_validators(self):
        validators = {}
        for arg in self.optional_args:
            validators[Optional(arg)] = self.arg_validator(arg)
        for arg in self.required_args:
            validators[arg] = self.arg_validator(arg)
        return validators

    def revalidate(self, arguments):
        if arguments.single_argument:
            if self.single_argument_allowed:
                arguments.validate_single_argument(self.arg_validator(self.single_argument_name))
            else:
                raise exceptions.StepMethodNeedsMoreThanOneArgument(
                    "Step method {0} requires {1} arguments, got one.".format(
                        self._method,
                        len(self.required_args),
                    )
                )
        else:
            if self._keywords:
                arguments.validate_kwargs(self.arg_validator(self._keywords))
            else:
                arguments.validate_args(self.mapping_validators)

    def method(self, arguments):
        if arguments.is_none:
            return self._method
        elif arguments.single_argument:
            if self.single_argument_allowed:
                return partial(self._method, **{self.single_argument_name: arguments.data})
        else:
            return partial(self._method, **arguments.data)
